#ifndef _DYNAMIXEL_MASTER_H
#define _DYNAMIXEL_MASTER_H

#include "comm.h"
#include "dynamixel.h"
#include "dynamixel2.h"


#ifndef MAX_DYN_MASTER_TX_BUFFER_LEN
  #define       MAX_DYN_MASTER_TX_BUFFER_LEN      256
#endif

#ifndef MAX_DYN_MASTER_RX_BUFFER_LEN
  #define       MAX_DYN_MASTER_RX_BUFFER_LEN      256
#endif

typedef struct
{
  TComm *comm_dev;
  TDynVersion version;
  unsigned short int op_length;
  unsigned char tx_buffer[MAX_DYN_MASTER_TX_BUFFER_LEN];
  unsigned char rx_buffer[MAX_DYN_MASTER_RX_BUFFER_LEN];
  TReturnLevel return_level;
  unsigned char packet_ready;
  unsigned char received_bytes;
  void (*set_tx_mode)(void);
  void (*set_rx_mode)(void);
  unsigned short int rx_timeout_ms;
  unsigned char rx_no_answer;
  unsigned char rx_num_packets;
}TDynamixelMaster;

/* public functions */
void dyn_master_init(TDynamixelMaster *master,TComm *dev,TDynVersion version);
void dyn_master_set_rx_timeout(TDynamixelMaster *master,unsigned short int timeout_ms);
void dyn_master_set_return_level(TDynamixelMaster *master,TReturnLevel level);
TReturnLevel dyn_master_get_return_level(TDynamixelMaster *master);
void dyn_master_scan(TDynamixelMaster *master,unsigned char *num,unsigned char *ids);
unsigned char dyn_master_ping(TDynamixelMaster *master,unsigned char id);
unsigned char dyn_master_read_byte(TDynamixelMaster *master,unsigned char id,unsigned short int address,unsigned char *data);
unsigned char dyn_master_read_word(TDynamixelMaster *master,unsigned char id,unsigned short int address,unsigned short int *data);
unsigned char dyn_master_read_table(TDynamixelMaster *master,unsigned char id,unsigned short int address,unsigned short int length,unsigned char *data);
unsigned char dyn_master_write_byte(TDynamixelMaster *master,unsigned char id, unsigned short int address, unsigned char data);
unsigned char dyn_master_write_word(TDynamixelMaster *master,unsigned char id, unsigned short int address, unsigned short int data);
unsigned char dyn_master_write_table(TDynamixelMaster *master,unsigned char id, unsigned short int address, unsigned short int length, unsigned char *data);
unsigned char dyn_master_reg_write(TDynamixelMaster *master,unsigned char id, unsigned short int address, unsigned short int length, unsigned char *data);
unsigned char dyn_master_action(TDynamixelMaster *master);
unsigned char dyn_master_sync_write(TDynamixelMaster *master,unsigned char num,unsigned char *ids,unsigned short int address, unsigned short int length, TWriteData *data);
unsigned char dyn_master_sync_read(TDynamixelMaster *master,unsigned char num,unsigned char *ids,unsigned short int address, unsigned short int length, TWriteData *data);
unsigned char dyn_master_bulk_read(TDynamixelMaster *master,unsigned char num,unsigned char *ids,unsigned short int *address, unsigned short int *length, TWriteData *data);
unsigned char dyn_master_bulk_write(TDynamixelMaster *master,unsigned char num,unsigned char *ids,unsigned short int *address, unsigned short int *length, TWriteData *data);
unsigned char dyn_master_relay(TDynamixelMaster *master,unsigned char *inst_pkt,unsigned char *status_pkt);

#endif
