#ifndef _DYNAMIXEL_SLAVE_H
#define _DYNAMIXEL_SLAVE_H

#include "comm.h"
#include "dynamixel.h"
#include "dynamixel2.h"

#ifndef MAX_DYN_SLAVE_TX_BUFFER_LEN
  #define       MAX_DYN_SLAVE_TX_BUFFER_LEN      256
#endif

#ifndef MAX_DYN_SLAVE_RX_BUFFER_LEN
  #define       MAX_DYN_SLAVE_RX_BUFFER_LEN      256
#endif

#ifndef MAX_DYN_SLAVE_REG_BUFFER_LEN
  #define       MAX_DYN_SLAVE_REG_BUFFER_LEN     256
#endif

typedef struct
{
  TComm *comm_dev;
  TDynVersion version;
  unsigned char address;
  unsigned char return_delay;
  TReturnLevel return_level;
  unsigned char packet_ready;
  unsigned char received_bytes;
  unsigned short int rx_timeout_ms;
  unsigned short int op_length;
  unsigned char tx_buffer[MAX_DYN_SLAVE_TX_BUFFER_LEN];
  unsigned char rx_buffer[MAX_DYN_SLAVE_RX_BUFFER_LEN];
  unsigned short int reg_address;
  unsigned short int reg_length;
  unsigned char reg_buffer[MAX_DYN_SLAVE_REG_BUFFER_LEN];
  unsigned short int sync_bulk_address;
  unsigned short int sync_bulk_length;
  unsigned char sync_bulk_prev_id;
  unsigned char sync_read_pending;
  unsigned char bulk_read_pending;
  void (*set_tx_mode)(void);
  void (*set_rx_mode)(void);
  void (*on_ping)(void);
  unsigned char (*on_read)(unsigned short int address,unsigned short int length,unsigned char *data);
  unsigned char (*on_write)(unsigned short int address,unsigned short int length,unsigned char *data);
  unsigned char (*on_reset)(void);
  unsigned char (*on_relay)(TDynVersion version,unsigned char *inst_pkt,unsigned char *status_pkt);
}TDynamixelSlave;

/* public functions */
void dyn_slave_init(TDynamixelSlave *slave,TComm *dev,unsigned char address,TDynVersion version);
void dyn_slave_set_rx_timeout(TDynamixelSlave *slave,unsigned short int timeout_ms);
void dyn_slave_set_address(TDynamixelSlave *slave,unsigned char address);
unsigned char dyn_slave_get_address(TDynamixelSlave *slave);
void dyn_slave_set_return_delay(TDynamixelSlave *slave,unsigned char delay);
unsigned char dyn_slave_get_return_delay(TDynamixelSlave *slave);
void dyn_slave_set_return_level(TDynamixelSlave *slave,TReturnLevel level);
TReturnLevel dyn_slave_get_return_level(TDynamixelSlave *slave);
TDynVersion dyn_slave_get_version(TDynamixelSlave *slave);
void dyn_slave_loop(TDynamixelSlave *slave);


#endif
