#ifndef _DYN_MANAGER_H
#define _DYN_MANAGER_H

#include "dynamixel_master.h"

#define    DYN_MANAGER_MAX_NUM_MASTERS       4
#define    DYN_MANAGER_MAX_NUM_MODULES       8
#define    DYN_MANAGER_MAX_NUM_DEVICES       32
#define    DYN_MANAGER_MAX_NUM_SINGLE_OP     16
#define    DYN_MANAGER_MAX_NUM_SYNC_OP       4
#define    DYN_MANAGER_MAX_NUM_BULK_OP       4
#define    DYN_MANAGER_MAX_NUM_OP            ((DYN_MANAGER_MAX_NUM_SINGLE_OP+DYN_MANAGER_MAX_NUM_SYNC_OP+DYN_MANAGER_MAX_NUM_BULK_OP)*DYN_MANAGER_MAX_NUM_MASTERS)

#define    DYN_MANAGER_PERIODIC              0xFF

#define    DYN_MANAGER_MODE_MASK             0x0F
#define    DYN_MANAGER_READ                  0x00
#define    DYN_MANAGER_WRITE                 0x01

#define    DYN_MANAGER_OP_MASK               0xF0
#define    DYN_MANAGER_SINGLE                0x10
#define    DYN_MANAGER_SYNC                  0x20
#define    DYN_MANAGER_BULK                  0x40
#define    DYN_MANAGER_NONE                  0x00

typedef enum {single_rd_op=DYN_MANAGER_SINGLE|DYN_MANAGER_READ,
              single_wr_op=DYN_MANAGER_SINGLE|DYN_MANAGER_WRITE,
              sync_rd_op=DYN_MANAGER_SYNC|DYN_MANAGER_READ,
              sync_wr_op=DYN_MANAGER_SYNC|DYN_MANAGER_WRITE,
              bulk_rd_op=DYN_MANAGER_BULK|DYN_MANAGER_READ,
              bulk_wr_op=DYN_MANAGER_BULK|DYN_MANAGER_WRITE,
              no_op=DYN_MANAGER_NONE} op_type_t;

typedef struct{
  op_type_t op_type;
  unsigned char op_index[DYN_MANAGER_MAX_NUM_MASTERS];
  unsigned char handle_index;
}OP_HANDLE;

typedef struct{
  op_type_t operation;
  unsigned char repetitions;
  unsigned char period_count;
  OP_HANDLE *op_handle;
}TDynManagerOpCommon;

typedef struct{
  TDynManagerOpCommon common;
  unsigned char id;
  unsigned short int address;
  unsigned short int length;
  unsigned char *data;
}TDynManagerSingleOp;

typedef struct{
  TDynManagerOpCommon common;
  unsigned char num_devices;
  unsigned char ids[DYN_MANAGER_MAX_NUM_DEVICES];
  unsigned short int address;
  unsigned short int length;
  TWriteData data[DYN_MANAGER_MAX_NUM_DEVICES];
}TDynManagerSyncOp;

typedef struct{
  TDynManagerOpCommon common;
  unsigned char num_devices;
  unsigned char ids[DYN_MANAGER_MAX_NUM_DEVICES];
  unsigned short int address[DYN_MANAGER_MAX_NUM_DEVICES];
  unsigned short int length[DYN_MANAGER_MAX_NUM_DEVICES];
  TWriteData data[DYN_MANAGER_MAX_NUM_DEVICES];
}TDynManagerBulkOp;

typedef struct{
  TDynManagerSingleOp single_op[DYN_MANAGER_MAX_NUM_SINGLE_OP];
  unsigned char num_single_op;
  TDynManagerSyncOp sync_op[DYN_MANAGER_MAX_NUM_SYNC_OP];
  unsigned char num_sync_op;
  TDynManagerBulkOp bulk_op[DYN_MANAGER_MAX_NUM_BULK_OP];
  unsigned char num_bulk_op;
}TDynManagerOps;

typedef struct{
  unsigned short model;
  unsigned char master_index;
}TDynManagerDevice;

struct TDynModule;

typedef struct{
  TDynamixelMaster *masters[DYN_MANAGER_MAX_NUM_MASTERS];
  TDynManagerOps operations[DYN_MANAGER_MAX_NUM_MASTERS];
  unsigned char num_masters;
  struct TDynModule *modules[DYN_MANAGER_MAX_NUM_MODULES];
  unsigned char num_modules;
  TDynManagerDevice devices[DYN_MANAGER_MAX_NUM_DEVICES];
  unsigned char num_devices;
  OP_HANDLE op_handles[DYN_MANAGER_MAX_NUM_OP];
  unsigned char num_ops;
}TDynManager;

// public functions
void dyn_manager_init(TDynManager *manager);
void dyn_manager_scan(TDynManager *manager);
void dyn_manager_add_master(TDynManager *manager,TDynamixelMaster *master);
unsigned char dyn_manager_get_num_masters(TDynManager *manager);
void dyn_manager_add_module(TDynManager *manager,struct TDynModule *module);
unsigned char dyn_manager_get_num_modules(TDynManager *manager);
void dyn_manager_delete_op(TDynManager *manager,OP_HANDLE *op);
void dyn_manager_set_op_period(TDynManager *manager,OP_HANDLE *op,unsigned char period_count);
unsigned char dyn_manager_get_op_period(TDynManager *manager,OP_HANDLE *op);
void dyn_manager_set_op_repetitions(TDynManager *manager,OP_HANDLE *op,unsigned char repetitions);
unsigned char dyn_manager_get_op_repetitions(TDynManager *manager,OP_HANDLE *op);
// single operation functions
OP_HANDLE *dyn_manager_single_op_new(TDynManager *manager,unsigned char mode,unsigned char id,unsigned short int address,unsigned short int length,unsigned char *data);
void dyn_manager_single_op_change_id(TDynManager *manager,OP_HANDLE *op,unsigned char id);
void dyn_manager_single_op_change_address(TDynManager *manager,OP_HANDLE *op,unsigned short int address);
void dyn_manager_single_op_change_length(TDynManager *manager,OP_HANDLE *op,unsigned short int length);
void dyn_manager_single_op_change_data(TDynManager *manager,OP_HANDLE *op,unsigned char *data);
// sync operation functions
OP_HANDLE *dyn_manager_sync_op_new(TDynManager *manager,unsigned char mode,unsigned char num,unsigned char *ids,unsigned short int address, unsigned short int length, TWriteData *data);
void dyn_manager_sync_op_add_devices(TDynManager *manager,OP_HANDLE *op,unsigned char num,unsigned char *ids,TWriteData *data);
void dyn_manager_sync_op_remove_devices(TDynManager *manager,OP_HANDLE *op,unsigned char num,unsigned char *ids);
void dyn_manager_sync_op_change_address(TDynManager *manager,OP_HANDLE *op,unsigned short int address);
void dyn_manager_sync_op_change_length(TDynManager *manager,OP_HANDLE *op,unsigned short int length);
void dyn_manager_sync_op_change_data(TDynManager *manager,OP_HANDLE *op,unsigned char num,unsigned char *ids,TWriteData *data);
// bulk operation functions
OP_HANDLE *dyn_manager_bulk_op_new(TDynManager *manager,unsigned char mode,unsigned char num,unsigned char *ids,unsigned short int *address, unsigned short int *length, TWriteData *data);
void dyn_manager_bulk_op_add_devices(TDynManager *manager,OP_HANDLE *op,unsigned char num,unsigned char *ids,unsigned short int *address, unsigned short int *length,TWriteData *data);
void dyn_manager_bulk_op_remove_devices(TDynManager *manager,OP_HANDLE *op,unsigned char num,unsigned char *ids);
void dyn_manager_bulk_op_change_address(TDynManager *manager,OP_HANDLE *op,unsigned char num,unsigned char *ids,unsigned short int *address);
void dyn_manager_bulk_op_change_length(TDynManager *manager,OP_HANDLE *op,unsigned char num,unsigned char *ids,unsigned short int *length);
void dyn_manager_bulk_op_change_data(TDynManager *manager,OP_HANDLE *op,unsigned char num,unsigned char *ids,TWriteData *data);
void dyn_manager_loop(TDynManager *manager);

#endif 
