#include <stdio.h>
#include <stdlib.h>
#include "dyn_devices.h"
#include "dyn_module.h"
#include "dyn_manager.h"

TDynamixelMaster dyn_master1;
TDynamixelMaster dyn_master2;
TDynManager manager;

// timer functions
unsigned long long int time_get_counts(void)
{
  static unsigned long long int count=0;
 
  count+=10000;
  return count;
}

unsigned int time_get_counts_per_us(void)
{
  return 10000;
}

// dynamixel master functions
unsigned char dyn_master_irq_receive_cb(void *dyn_master,unsigned char byte)
{

}

unsigned char dyn_master_dma_send_cb(void *dyn_master)
{

}

unsigned char dyn_master_dma_receive_cb(void *dyn_master)
{

}

void dummy_dyn_master_set_tx_mode(void)
{

}

void dummy_dyn_master_set_rx_mode(void)
{

}

void dyn_master_init(TDynamixelMaster *master,TComm *dev,TDynVersion version)
{
  /* assign communication functions */
  dev->irq_receive_cb=dyn_master_irq_receive_cb;
  dev->dma_send_cb=dyn_master_dma_send_cb;
  dev->dma_receive_cb=dyn_master_dma_receive_cb;
  master->comm_dev=dev;
  master->version=version;
  dev->data=master;
  /* initialize the internal callbacks */
  master->set_tx_mode=dummy_dyn_master_set_tx_mode;
  master->set_rx_mode=dummy_dyn_master_set_rx_mode;
  /* initialize internal variables */
  master->packet_ready=0x00;
  master->received_bytes=0x00;
  master->rx_timeout_ms=50;
  master->rx_no_answer=0x00;
  master->rx_num_packets=0x00;
  master->return_level=return_all;

  master->set_rx_mode();
}

void dyn_master_scan(TDynamixelMaster *master,unsigned char *num,unsigned char *ids)
{
  unsigned int i;

  if(master==&dyn_master1)
  {
    for(i=0;i<20;i++)
      ids[i]=i+1;
    (*num)=20;
  }
  else if(master==&dyn_master2)
  {
    for(i=0;i<4;i++)
      ids[i]=i+21;
    (*num)=4;
  }
}

unsigned char dyn_master_read_word(TDynamixelMaster *master,unsigned char id,unsigned short int address,unsigned short int *data)
{
  if(master==&dyn_master1)
  {
    if(id>0 && id<11)
    {
      if(address==0x0000)
        (*data)=SERVO_AX12A;
    } 
    if(id>10 && id<21)
    {
      if(address==0x0000)
        (*data)=SERVO_MX28;
    } 
  }
  else if(master==&dyn_master2)
  {
    if(id>20 && id<25)
    {
      if(address==0x0000)
        (*data)=SERVO_XL320;
    } 
  }
}

unsigned char dyn_master_start_read_table(TDynamixelMaster *master,unsigned char id,unsigned short int address,unsigned short int length,unsigned char *data)
{
  if(master==&dyn_master1)
    printf("Master 1\n");
  else
    printf("Master 2\n");
  printf("  read operation to servo %d at address %d with length %d\n",id,address,length);

  return DYN_SUCCESS;
}

unsigned char dyn_master_is_read_table_done(TDynamixelMaster *master,unsigned char id,unsigned short int address,unsigned short int length,unsigned char *data)
{
  return DYN_SUCCESS;
}

unsigned char dyn_master_start_write_table(TDynamixelMaster *master,unsigned char id,unsigned short int address,unsigned short int length,unsigned char *data)
{
  unsigned short int i;

  if(master==&dyn_master1)
    printf("Master 1\n");
  else
    printf("Master 2\n");
  printf("  write operation to servo %d at address %d with length %d\n",id,address,length);
  printf("  data:\n");
  for(i=0;i<length;i++)
    printf("    0x%x\n",data[i]);

  return DYN_SUCCESS;
}

unsigned char dyn_master_is_write_table_done(TDynamixelMaster *master,unsigned char id,unsigned short int address,unsigned short int length,unsigned char *data)
{
  return DYN_SUCCESS;
}

unsigned char dyn_master_start_sync_write(TDynamixelMaster *master,unsigned char num,unsigned char *ids,unsigned short int address, unsigned short int length, TWriteData *data)
{
  unsigned short int i,j;

  if(master==&dyn_master1)
    printf("Master 1\n");
  else
    printf("Master 2\n");
  printf("  sync write operation at address %d with length %d\n",address,length);
  for(j=0;j<num;j++)
  {
    printf("  servo %d data:\n",ids[j]);
    for(i=0;i<length;i++)
      printf("    0x%x\n",data[j].data_addr[i]);
  }

  return DYN_SUCCESS;
}

unsigned char dyn_master_is_sync_write_done(TDynamixelMaster *master,unsigned char num,unsigned char *ids,unsigned short int address, unsigned short int length, TWriteData *data)
{
  return DYN_SUCCESS;
}

unsigned char dyn_master_start_sync_read(TDynamixelMaster *master,unsigned char num,unsigned char *ids,unsigned short int address, unsigned short int length, TWriteData *data)
{
  unsigned short int i;

  if(master==&dyn_master1)
    printf("Master 1\n");
  else
    printf("Master 2\n");
  printf("  sync read operation at address %d with length %d for servos:\n",address,length);
  for(i=0;i<num;i++)
    printf("  servo %d:\n",ids[i]);

  return DYN_SUCCESS;
}

unsigned char dyn_master_is_sync_read_done(TDynamixelMaster *master,unsigned char num,unsigned char *ids,unsigned short int address, unsigned short int length, TWriteData *data)
{
  return DYN_SUCCESS;
}

unsigned char dyn_master_start_bulk_write(TDynamixelMaster *master,unsigned char num,unsigned char *ids,unsigned short int *address, unsigned short int *length, TWriteData *data)
{
  unsigned short int i,j;

  if(master==&dyn_master1)
    printf("Master 1\n");
  else
    printf("Master 2\n");
  printf("  bulk write operation:\n");
  for(j=0;j<num;j++)
  {
    printf("  servo %d at address %d with length %d and data:\n",ids[j],address[j],length[j]);
    for(i=0;i<length[j];i++)
      printf("    0x%x\n",data[j].data_addr[i]);
  }

  return DYN_SUCCESS;
}

unsigned char dyn_master_is_bulk_write_done(TDynamixelMaster *master,unsigned char num,unsigned char *ids,unsigned short int *address, unsigned short int *length, TWriteData *data)
{
  return DYN_SUCCESS;
}

unsigned char dyn_master_start_bulk_read(TDynamixelMaster *master,unsigned char num,unsigned char *ids,unsigned short int *address, unsigned short int *length, TWriteData *data)
{
  unsigned short int i,j;

  if(master==&dyn_master1)
    printf("Master 1\n");
  else
    printf("Master 2\n");
  printf("  bulk read operation:\n");
  for(j=0;j<num;j++)
    printf("  servo %d at address %d with length %d\n",ids[j],address[j],length[j]);

  return DYN_SUCCESS;
}

unsigned char dyn_master_is_bulk_read_done(TDynamixelMaster *master,unsigned char num,unsigned char *ids,unsigned short int *address, unsigned short int *length, TWriteData *data)
{
  return DYN_SUCCESS;
}

// module functions
unsigned char single_op_data[8];
unsigned char sync_op_data[64];
unsigned char bulk_op_data[64];

void module1_init(void)
{
  unsigned int i;

  printf("module1: init function\n");
  for(i=0;i<8;i++)
    single_op_data[i]=i;
  for(i=0;i<64;i++)
  {
    sync_op_data[i]=i;
    bulk_op_data[i]=i;
  }
}

void module1_add_device(unsigned char id,unsigned short int model)
{
  printf("module1: add_device function\n");
  printf("  ID: %d model: %d\n",id,model);
}

void module1_setup(void)
{
  TWriteData data[8];
  unsigned char servo_ids[8];
  unsigned short int address[8],length[8];

  servo_ids[0]=12;
  servo_ids[1]=21;

  data[0].data_addr=&sync_op_data[4];
  data[1].data_addr=&sync_op_data[8];

  printf("module1: setup function\n");
  dyn_manager_single_op_new(&manager,DYN_MANAGER_WRITE,11,0x0000,2,&single_op_data[0]);
  dyn_manager_single_op_new(&manager,DYN_MANAGER_READ,11,0x0020,4,&single_op_data[0]);
  dyn_manager_single_op_new(&manager,DYN_MANAGER_WRITE,21,0x0050,6,&single_op_data[0]);
  dyn_manager_single_op_new(&manager,DYN_MANAGER_READ,22,0x0070,8,&single_op_data[0]);
  dyn_manager_sync_op_new(&manager,DYN_MANAGER_WRITE,2,servo_ids,0x0000,2,data);
  dyn_manager_sync_op_new(&manager,DYN_MANAGER_READ,2,servo_ids,0x0000,2,data);
  servo_ids[0]=12;
  servo_ids[1]=13;
  servo_ids[2]=14;
  servo_ids[3]=15;
  servo_ids[4]=21;
  servo_ids[5]=22;
  servo_ids[6]=23;
  servo_ids[7]=24;
  data[0].data_addr=&sync_op_data[0];
  address[0]=0;
  length[0]=1;
  data[1].data_addr=&sync_op_data[1];
  address[1]=1;
  length[1]=2;
  data[2].data_addr=&sync_op_data[3];
  address[2]=3;
  length[2]=4;
  data[3].data_addr=&sync_op_data[7];
  address[3]=7;
  length[3]=8;
  data[4].data_addr=&sync_op_data[15];
  address[4]=15;
  length[4]=8;
  data[5].data_addr=&sync_op_data[23];
  address[5]=23;
  length[5]=4;
  data[6].data_addr=&sync_op_data[27];
  address[6]=27;
  length[6]=2;
  data[7].data_addr=&sync_op_data[29];
  address[7]=29;
  length[7]=1;
  dyn_manager_bulk_op_new(&manager,DYN_MANAGER_WRITE,8,servo_ids,address,length,data);
  dyn_manager_bulk_op_new(&manager,DYN_MANAGER_READ,8,servo_ids,address,length,data);
}

void module1_pre_process(void)
{
  printf("module1: pre_process function\n");
}

void module1_post_process(void)
{
  printf("module1: post_process function\n");
}

void module2_init(void)
{
  printf("module2: init function\n");
}

void module2_add_device(unsigned char id,unsigned short int model)
{
  printf("module2: add_device function\n");
  printf("  ID: %d model: %d\n",id,model);
}

void module2_setup(void)
{
  printf("module2: setup function\n");
  dyn_manager_single_op_new(&manager,DYN_MANAGER_WRITE,1,0x0050,6,&single_op_data[0]);
  dyn_manager_single_op_new(&manager,DYN_MANAGER_READ,2,0x0070,8,&single_op_data[0]);
}

void module2_pre_process(void)
{
  printf("module2: pre_process function\n");
}

void module2_post_process(void)
{
  printf("module2: post_process function\n");
}

int main(void)
{
  int i;
  TDynModule module1;
  TDynModule module2;
  TTime dyn_master1_timer; 
  TComm dyn_master1_comm;
  TTime dyn_master2_timer; 
  TComm dyn_master2_comm;

  // init modules
  dyn_module_init(&module1);
  dyn_module_add_model(&module1,SERVO_MX28);
  dyn_module_add_model(&module1,SERVO_MX28);// this should be ignored
  dyn_module_add_model(&module1,SERVO_XL320);
  dyn_module_set_init_fctn(&module1,module1_init);
  dyn_module_set_add_device_fctn(&module1,module1_add_device);
  dyn_module_set_setup_fctn(&module1,module1_setup);
  dyn_module_set_pre_process_fctn(&module1,module1_pre_process);
  dyn_module_set_post_process_fctn(&module1,module1_post_process);
  dyn_module_set_period(&module1,1);

  dyn_module_init(&module2);
  dyn_module_add_model(&module2,SERVO_AX12A);
  dyn_module_set_init_fctn(&module2,module1_init);
  dyn_module_set_add_device_fctn(&module2,module1_add_device);
  dyn_module_set_setup_fctn(&module2,module1_setup);
  dyn_module_set_pre_process_fctn(&module2,module1_pre_process);
  dyn_module_set_post_process_fctn(&module2,module1_post_process);
  dyn_module_set_period(&module2,1);

  // init masters
  time_init(&dyn_master1_timer,time_get_counts_per_us(),time_get_counts);
  comm_init(&dyn_master1_comm,0x01,&dyn_master1_timer);
  dyn_master_init(&dyn_master1,&dyn_master1_comm,DYN_VER1);
  time_init(&dyn_master2_timer,time_get_counts_per_us(),time_get_counts);
  comm_init(&dyn_master2_comm,0x01,&dyn_master2_timer);
  dyn_master_init(&dyn_master2,&dyn_master2_comm,DYN_VER2);

  dyn_manager_init(&manager);
  dyn_manager_add_master(&manager,&dyn_master1);
  dyn_manager_add_master(&manager,&dyn_master2);
  dyn_manager_add_module(&manager,&module1);
  dyn_manager_add_module(&manager,&module2);
  dyn_manager_scan(&manager);

  for(i=0;i<10;i++)
    dyn_manager_loop(&manager);
}
