#include "usart1.h"

#define     USART                    USART1
#define     USART_ENABLE_CLK         __USART1_CLK_ENABLE()
#define     USART_IRQn               USART1_IRQn
#define     USART_IRQHandler         USART1_IRQHandler

#define     USART_TX_PIN             GPIO_PIN_9                
#define     USART_TX_GPIO_PORT       GPIOA
#define     USART_ENABLE_TX_GPIO_CLK __GPIOA_CLK_ENABLE()

#define     USART_RX_PIN             GPIO_PIN_10
#define     USART_RX_GPIO_PORT       GPIOA
#define     USART_ENABLE_RX_GPIO_CLK __GPIOA_CLK_ENABLE()

/* DMA configuration */
#define     USART_DMA                DMA1
#define     USART_ENABLE_DMA_CLK     __DMA1_CLK_ENABLE()

#define     USART_TX_DMA_CHANNEL     DMA1_Channel2
#define     USART_TX_DMA_REQUEST     DMA_REQUEST_3
#define     USART_RX_DMA_CHANNEL     DMA1_Channel3
#define     USART_RX_DMA_REQUEST     DMA_REQUEST_3

#define     USART_DMA_IRQn           DMA1_Channel2_3_IRQn
#define     USART_DMA_IRQHandler     DMA1_Channel2_3_IRQHandler

// private variables
UART_HandleTypeDef UartHandle;
DMA_HandleTypeDef hdma_tx;
DMA_HandleTypeDef hdma_rx;
TComm *usart1_comm_dev;

// interrupt handlers
void USART_IRQHandler(void)
{
  unsigned char data,ret;
  uint32_t source;

  if(__HAL_UART_GET_IT(&UartHandle, UART_IT_RXNE) != RESET)
  {
    if(__HAL_UART_GET_IT_SOURCE(&UartHandle, UART_IT_RXNE) !=RESET)
    {
      __HAL_UART_CLEAR_IT(&UartHandle,UART_FLAG_RXNE);
      data=(uint8_t)(UartHandle.Instance->RDR & (uint8_t)0x00FF);
      // call the reception function
      if(!comm_do_irq_receive(usart1_comm_dev,data))
        __HAL_UART_DISABLE_IT(&UartHandle, UART_IT_RXNE);
    }
  }
  if(__HAL_UART_GET_IT(&UartHandle, UART_IT_TC) != RESET)
  {
    if(__HAL_UART_GET_IT_SOURCE(&UartHandle, UART_IT_TC) !=RESET)
    {
      __HAL_UART_CLEAR_IT(&UartHandle,UART_FLAG_TC);
      ret=comm_do_irq_send(usart1_comm_dev,&data);
      if(ret==0x01)
        UartHandle.Instance->TDR=data;
      else if(ret==0x00)
        __HAL_UART_DISABLE_IT(&UartHandle, UART_IT_TC);
    }
  }
  /* check errors */
  if(__HAL_UART_GET_IT(&UartHandle, UART_IT_PE) != RESET)/* parity error */
  {
    if(__HAL_UART_GET_IT_SOURCE(&UartHandle, UART_IT_PE) != RESET)
    {
      __HAL_UART_CLEAR_IT(&UartHandle, UART_CLEAR_PEF);
    }
  }
  source=__HAL_UART_GET_IT_SOURCE(&UartHandle, UART_IT_ERR);
  if(__HAL_UART_GET_IT(&UartHandle, UART_IT_FE) != RESET)/* framing error */
  {
    if(source != RESET)
    {
      __HAL_UART_CLEAR_IT(&UartHandle, UART_CLEAR_FEF);
    }
  }
  if(__HAL_UART_GET_IT(&UartHandle, UART_IT_NE) != RESET)/* noise error */
  {
    if(source != RESET)
    {
      __HAL_UART_CLEAR_IT(&UartHandle, UART_CLEAR_NEF);
    }
  }
  if(__HAL_UART_GET_IT(&UartHandle, UART_IT_ORE) != RESET)/* overrun error */
  {
    if(source != RESET)
    {
      __HAL_UART_CLEAR_IT(&UartHandle, UART_CLEAR_OREF);
    }
  }
}

void USART_DMA_IRQHandler(void)
{
  if(__HAL_DMA_GET_FLAG(UartHandle.hdmatx,__HAL_DMA_GET_TE_FLAG_INDEX(UartHandle.hdmatx)) != RESET)
  {
    if(__HAL_DMA_GET_IT_SOURCE(UartHandle.hdmatx, DMA_IT_TE) != RESET)
    {
      /* Disable the transfer error interrupt */
      __HAL_DMA_DISABLE_IT(UartHandle.hdmatx, DMA_IT_TE);
      /* Clear the transfer error flag */
      __HAL_DMA_CLEAR_FLAG(UartHandle.hdmatx, __HAL_DMA_GET_TE_FLAG_INDEX(UartHandle.hdmatx));
    }
  }
  if(__HAL_DMA_GET_FLAG(UartHandle.hdmatx,__HAL_DMA_GET_TC_FLAG_INDEX(UartHandle.hdmatx)) != RESET)
  {
    if(__HAL_DMA_GET_IT_SOURCE(UartHandle.hdmatx, DMA_IT_TC) != RESET)
    {
      /* Disable the half transfer interrupt */
      __HAL_DMA_DISABLE_IT(UartHandle.hdmatx, DMA_IT_TC);
      /* Clear the transfer complete flag */
      __HAL_DMA_CLEAR_FLAG(UartHandle.hdmatx, __HAL_DMA_GET_TC_FLAG_INDEX(UartHandle.hdmatx));
      CLEAR_BIT(UartHandle.Instance->CR3, (USART_CR3_DMAT | USART_CR3_DMAR));
      HAL_DMA_Abort(UartHandle.hdmatx);
      // call the user function
      comm_do_dma_send(usart1_comm_dev);
    }
  }
  if(__HAL_DMA_GET_FLAG(UartHandle.hdmatx, __HAL_DMA_GET_HT_FLAG_INDEX(UartHandle.hdmatx)) != RESET)
  {
    if(__HAL_DMA_GET_IT_SOURCE(UartHandle.hdmatx, DMA_IT_HT) != RESET)
    {
      /* Disable the half transfer interrupt */
      __HAL_DMA_DISABLE_IT(UartHandle.hdmatx, DMA_IT_HT);
      /* Clear the half transfer complete flag */
      __HAL_DMA_CLEAR_FLAG(UartHandle.hdmatx, __HAL_DMA_GET_HT_FLAG_INDEX(UartHandle.hdmatx));
    }
  }
  if(__HAL_DMA_GET_FLAG(UartHandle.hdmarx,__HAL_DMA_GET_TE_FLAG_INDEX(UartHandle.hdmarx)) != RESET)
  {
    if(__HAL_DMA_GET_IT_SOURCE(UartHandle.hdmarx, DMA_IT_TE) != RESET)
    {
      /* Disable the transfer error interrupt */
      __HAL_DMA_DISABLE_IT(UartHandle.hdmarx, DMA_IT_TE);
      /* Clear the transfer error flag */
      __HAL_DMA_CLEAR_FLAG(UartHandle.hdmarx, __HAL_DMA_GET_TE_FLAG_INDEX(UartHandle.hdmarx));
    }
  }
  if(__HAL_DMA_GET_FLAG(UartHandle.hdmarx,__HAL_DMA_GET_TC_FLAG_INDEX(UartHandle.hdmarx)) != RESET)
  {
    if(__HAL_DMA_GET_IT_SOURCE(UartHandle.hdmarx, DMA_IT_TC) != RESET)
    {
      /* Disable the transfer complete interrupt */
      __HAL_DMA_DISABLE_IT(UartHandle.hdmarx, DMA_IT_TC);
      /* Clear the transfer complete flag */
      __HAL_DMA_CLEAR_FLAG(UartHandle.hdmarx, __HAL_DMA_GET_TC_FLAG_INDEX(UartHandle.hdmarx));
      CLEAR_BIT(UartHandle.Instance->CR3, (USART_CR3_DMAT | USART_CR3_DMAR));
      HAL_DMA_Abort(UartHandle.hdmarx);
      // call the user function
      comm_do_dma_receive(usart1_comm_dev);
    }
  }
  if(__HAL_DMA_GET_FLAG(UartHandle.hdmarx, __HAL_DMA_GET_HT_FLAG_INDEX(UartHandle.hdmarx)) != RESET)
  {
    if(__HAL_DMA_GET_IT_SOURCE(UartHandle.hdmarx, DMA_IT_HT) != RESET)
    {
      /* Disable the half transfer interrupt */
      __HAL_DMA_DISABLE_IT(UartHandle.hdmarx, DMA_IT_HT);
      /* Clear the half transfer complete flag */
      __HAL_DMA_CLEAR_FLAG(UartHandle.hdmarx, __HAL_DMA_GET_HT_FLAG_INDEX(UartHandle.hdmarx));
    }
  }
}

/* public functions*/
void usart1_init(TComm *comm_dev,UART_InitTypeDef *conf,UART_AdvFeatureInitTypeDef *adv_conf,TUSART_IRQ_Priorities *priorities)
{
  GPIO_InitTypeDef GPIO_InitStructure;

  /* Enable GPIO clock */
  USART_ENABLE_TX_GPIO_CLK;
  USART_ENABLE_RX_GPIO_CLK;
  USART_ENABLE_DMA_CLK;
  // configure the GPIO pins

  /* Configure USART Tx and Rx as alternate function push-pull */
  GPIO_InitStructure.Pin       = USART_TX_PIN;
  GPIO_InitStructure.Mode      = GPIO_MODE_AF_PP;
  GPIO_InitStructure.Pull      = GPIO_PULLUP;
  GPIO_InitStructure.Speed     = GPIO_SPEED_HIGH;
  GPIO_InitStructure.Alternate = GPIO_AF4_USART1;
  HAL_GPIO_Init(USART_TX_GPIO_PORT, &GPIO_InitStructure);

  GPIO_InitStructure.Pin       = USART_RX_PIN;
  HAL_GPIO_Init(USART_RX_GPIO_PORT, &GPIO_InitStructure);

  USART_ENABLE_CLK;

  UartHandle.Instance = USART;
  usart1_config(comm_dev,conf,adv_conf);

  if(comm_dev->use_dma)
  {
    // configure the DMA channels
    hdma_tx.Instance                 = USART_TX_DMA_CHANNEL;
    hdma_tx.Init.Request             = USART_TX_DMA_REQUEST;
    hdma_tx.Init.Direction           = DMA_MEMORY_TO_PERIPH;
    hdma_tx.Init.PeriphInc           = DMA_PINC_DISABLE;
    hdma_tx.Init.MemInc              = DMA_MINC_ENABLE;
    hdma_tx.Init.PeriphDataAlignment = DMA_PDATAALIGN_BYTE;
    hdma_tx.Init.MemDataAlignment    = DMA_MDATAALIGN_BYTE;
    hdma_tx.Init.Mode                = DMA_NORMAL;
    hdma_tx.Init.Priority            = DMA_PRIORITY_LOW;

    HAL_DMA_Init(&hdma_tx);

    // Associate the initialized DMA handle to the UART handle 
    __HAL_LINKDMA(&UartHandle, hdmatx, hdma_tx);

    // Configure the DMA handler for reception process 
    hdma_rx.Instance                 = USART_RX_DMA_CHANNEL;
    hdma_rx.Init.Request             = USART_RX_DMA_REQUEST;
    hdma_rx.Init.Direction           = DMA_PERIPH_TO_MEMORY;
    hdma_rx.Init.PeriphInc           = DMA_PINC_DISABLE;
    hdma_rx.Init.MemInc              = DMA_MINC_ENABLE;
    hdma_rx.Init.PeriphDataAlignment = DMA_PDATAALIGN_BYTE;
    hdma_rx.Init.MemDataAlignment    = DMA_MDATAALIGN_BYTE;
    hdma_rx.Init.Mode                = DMA_NORMAL;
    hdma_rx.Init.Priority            = DMA_PRIORITY_HIGH;

    HAL_DMA_Init(&hdma_rx);

    // Associate the initialized DMA handle to the the UART handle
    __HAL_LINKDMA(&UartHandle, hdmarx, hdma_rx);
  }
  usart1_set_priorities(comm_dev,priorities);

  /* Initialize the comm structure */
  comm_dev->send_irq=usart1_send_irq;
  comm_dev->enable_tx_irq=usart1_enable_tx_irq;
  comm_dev->receive_irq=usart1_receive_irq;
  comm_dev->cancel_receive_irq=usart1_cancel_receive_irq;
  if(comm_dev->use_dma)
  {
    comm_dev->send_dma=usart1_send_dma;
    comm_dev->receive_dma=usart1_receive_dma;
    comm_dev->cancel_receive_dma=usart1_cancel_receive_dma;
  }
  comm_dev->irq_send_cb=0x00000000;
  comm_dev->irq_receive_cb=0x00000000;
  comm_dev->dma_send_cb=0x00000000;
  comm_dev->dma_receive_cb=0x00000000;

  /* initialize internal variables */
  usart1_comm_dev=comm_dev;
}

void usart1_config(TComm *comm_dev,UART_InitTypeDef *conf,UART_AdvFeatureInitTypeDef *adv_conf)
{
  UartHandle.Init.BaudRate     = conf->BaudRate;
  UartHandle.Init.WordLength   = conf->WordLength;
  UartHandle.Init.StopBits     = conf->StopBits;
  UartHandle.Init.Parity       = conf->Parity;
  UartHandle.Init.Mode         = conf->Mode;
  UartHandle.Init.HwFlowCtl    = conf->HwFlowCtl;
  UartHandle.Init.OverSampling = conf->OverSampling;
  UartHandle.Init.OneBitSampling = conf->OneBitSampling;
  UartHandle.AdvancedInit.AdvFeatureInit = adv_conf->AdvFeatureInit;
  UartHandle.AdvancedInit.TxPinLevelInvert = adv_conf->TxPinLevelInvert;
  UartHandle.AdvancedInit.RxPinLevelInvert = adv_conf->RxPinLevelInvert;
  UartHandle.AdvancedInit.DataInvert = adv_conf->DataInvert;
  UartHandle.AdvancedInit.Swap = adv_conf->Swap;
  UartHandle.AdvancedInit.OverrunDisable = adv_conf->OverrunDisable;
  UartHandle.AdvancedInit.DMADisableonRxError = adv_conf->DMADisableonRxError;
  UartHandle.AdvancedInit.AutoBaudRateEnable = adv_conf->AutoBaudRateEnable;
  UartHandle.AdvancedInit.AutoBaudRateMode = adv_conf->AutoBaudRateMode;
  UartHandle.AdvancedInit.MSBFirst = adv_conf->MSBFirst;
  HAL_UART_Init(&UartHandle);
}

void usart1_set_priorities(TComm *comm_dev,TUSART_IRQ_Priorities *priorities)
{
  HAL_NVIC_SetPriority(USART_IRQn, priorities->irq_priority,priorities->irq_subpriority);
  HAL_NVIC_EnableIRQ(USART_IRQn);
  if(comm_dev->use_dma)
  {
    HAL_NVIC_SetPriority(USART_DMA_IRQn, priorities->dma_priority,priorities->dma_subpriority);
    HAL_NVIC_EnableIRQ(USART_DMA_IRQn);
  }
}

/* IRQ functions */
unsigned char usart1_send_irq(unsigned char first_byte)
{
  __HAL_UART_CLEAR_IT(&UartHandle,UART_FLAG_TC);
  __HAL_UART_ENABLE_IT(&UartHandle, UART_IT_TC);
  UartHandle.Instance->TDR=first_byte;

  return 0x00;
}

unsigned char usart1_enable_tx_irq(void)
{
  __HAL_UART_CLEAR_IT(&UartHandle,UART_FLAG_TC);
  __HAL_UART_ENABLE_IT(&UartHandle, UART_IT_TC);

  return 0x00;
}

unsigned char usart1_receive_irq(void)
{
  /* enable the rx interrupt */
  __HAL_UART_ENABLE_IT(&UartHandle, UART_IT_RXNE);

  return 0x00;
}

unsigned char usart1_cancel_receive_irq(void)
{
  /* disable the rx interrupt */
  __HAL_UART_DISABLE_IT(&UartHandle, UART_IT_RXNE);

  return 0x00;
}

/* DMA functions */
unsigned char usart1_send_dma(unsigned char *data,unsigned short int length)
{
  HAL_DMA_Start_IT(UartHandle.hdmatx,(uint32_t)data,(uint32_t)&UartHandle.Instance->TDR,length);
  /* Clear the TC flag in the SR register by writing 0 to it */
  __HAL_UART_CLEAR_IT(&UartHandle,UART_FLAG_TC);
  /* Enable the DMA transfer for transmit request by setting the DMAT bit
     in the UART CR3 register */
  SET_BIT(UartHandle.Instance->CR3, USART_CR3_DMAT);

  return 0x00;
}

unsigned char usart1_receive_dma(unsigned char *data,unsigned short int length)
{
  HAL_DMA_Start_IT(UartHandle.hdmarx,(uint32_t)&UartHandle.Instance->RDR,(uint32_t)data,length);
  SET_BIT(UartHandle.Instance->CR3, USART_CR3_DMAR);

  return 0x00;
}

unsigned char usart1_cancel_receive_dma(void)
{
  CLEAR_BIT(UartHandle.Instance->CR3, (USART_CR3_DMAT | USART_CR3_DMAR));
  HAL_DMA_Abort(UartHandle.hdmarx);

  return 0x00;
}

